import hashlib
import re
from collections import OrderedDict
from os.path import dirname, exists, isfile, join
from pathlib import Path

from pyorbs.templates import render


class Requirements:
    """
    Dynamic requirements file representation.

    Attributes:
        path: The path to the requirements file.
        locked: The path to the requirements lockfile.
        changed: Whether the requirements lockfile is up-to-date.

    """
    def __init__(self, path, bare=False):
        """
        Args:
            path (str): The path to the requirements file.
            bare (bool): Whether to use the bare requirements file.

        """
        if not exists(path) or not isfile(path):
            raise ValueError('Invalid requirements file "%s"' % path)
        self.path = path
        self.locked = path + '.lock'
        self._bare = bare
        self._hash, self._options = self._lockfile_context() if not bare else (None, None)
        self.changed = not bare and self._hash != self._stored_hash()

    def __str__(self):
        """
        Returns the path to the relevant requirements file.
        """
        return self.path if self._bare or self.changed else self.locked

    def lock(self, frozen):
        """
        Updates the lockfile using the provided frozen requirements.
        """
        header = render('lockfile_header', {'reqs': self.path, 'hash': self._hash})
        Path(self.locked).write_text(header + '\n'.join(self._options + [frozen]))
        print('Frozen requirements are written to "%s"' % self.locked)

    def _lockfile_context(self):
        """
        Returns a tuple with the SHA-256 hash of the concatenated requirements files and the list
        of additional options used in the requirements files.
        """
        hash_value = hashlib.sha256()
        options = []
        done = OrderedDict([(self.path, False)])
        while not all(done.values()):
            reqs = [r for r, p in done.items() if not p][0]
            if not exists(reqs):
                raise RuntimeError('Requirements file "%s" not found (referenced by "%s")' %
                                   (reqs, self.path))
            text = Path(reqs).read_text()
            hash_value.update(text.encode())
            options += re.findall(r'^(-[^rc].*)$', text, re.MULTILINE)
            done.update([(join(dirname(reqs), r), False)
                         for r in re.findall(r'^-[rc] (.*)$', text, re.MULTILINE)
                         if join(dirname(reqs), r) not in done])
            done[reqs] = True
        return hash_value.hexdigest(), options

    def _stored_hash(self):
        """
        Returns the stored hash from the lockfile or None if no lockfile or hash is found.
        """
        if not exists(self.locked):
            return None
        hash_search = re.search(r'hash: (.*)', Path(self.locked).read_text())
        return hash_search.group(1) if hash_search else None



import sys
import tempfile
from os import environ, getcwd, remove, walk
from os.path import exists, isdir, join, realpath
from pathlib import Path
from shutil import rmtree

from pyorbs.reqs import Requirements
from pyorbs.shell import SHELLS, current_shell_type, execute, which
from pyorbs.templates import render


class Orbs:
    def __init__(self, path):
        self.path = path
        self.orbs = next(walk(path))[1] if exists(path) else []
        self._glowing_file = join(self.path, '.glowing')

    def list(self):
        orbs = [orb + ' *' if orb == self.glowing() else orb for orb in self.orbs]
        print('\n'.join(sorted(orbs)) if orbs else 'There are no orbs')

    @staticmethod
    def freeze(reqs, executable=sys.executable):
        """
        Freeze the requirements files in a given path.

        Args:
            reqs (str): The path to the requirements files.
            executable (str): The Python executable to use.

        """
        for reqs_obj in Orbs.reqs_list(reqs):
            if reqs_obj.changed:
                print('Freezing requirements "%s"...' % reqs_obj.path)
                with tempfile.TemporaryDirectory(prefix='pyorbs-') as tmp_path:
                    orb = Orbs(tmp_path).orb(name='frozen', new=True)
                    orb.make(reqs_obj, executable, quiet=True, update=True)
            else:
                print('Requirements lockfile of "%s" is up-to-date' % reqs_obj.path)

    @staticmethod
    def test(reqs, quiet=False):
        """
        Returns 0 when all requirements files are up-to-date in a given path and 1 otherwise.

        Args:
            reqs (str): The path to the requirements files.
            quiet (bool): Whether to suppress info messages.

        """
        outdated = False
        for reqs_obj in Orbs.reqs_list(reqs):
            if reqs_obj.changed:
                outdated = True
                if not quiet:
                    print('Requirements lockfile of "%s" is outdated' % reqs_obj.path)
            elif not quiet:
                print('Requirements lockfile of "%s" is up-to-date' % reqs_obj.path)
        return int(outdated)

    @staticmethod
    def reqs_list(reqs):
        if not exists(reqs):
            raise ValueError('Requirements path "%s" not found' % reqs)
        if isdir(reqs):
            reqs_list = [Requirements(join(reqs, reqs_file)) for reqs_file in next(walk(reqs))[2]
                         if not (reqs_file.endswith('.lock') or reqs_file.startswith('.'))]
        else:
            reqs_list = [Requirements(reqs)]
        if not reqs_list:
            raise ValueError('There are no requirements files in path "%s"' % reqs)
        return reqs_list

    def glowing(self):
        return Path(self._glowing_file).read_text() if exists(self._glowing_file) else None

    def toggle_glow(self, name=None, force_on=False):
        """
        Toggle orb glow.

        Args:
            name (str): The name of the orb. Defaults to the currently active orb.
            force_on (bool): Whether to force orb glow to be turned on.

        """
        name = name or environ.get('PYORBS_ACTIVE_ORB', None)
        if (not name or name == self.glowing()) and not force_on:
            if exists(self._glowing_file):
                remove(self._glowing_file)
            print('No orb shall glow now')
        elif name not in self.orbs:
            raise ValueError('Invalid orb name "%s"' % name)
        else:
            Path(self._glowing_file).write_text(name)
            print('Orb "%s" is glowing now' % name)

    def orb(self, name=None, new=False, shell=False):
        """
        Create an Orb instance.

        Args:
            name (str): The name of the orb. Defaults to the name of the glowing orb.
            new (bool): Whether the orb is new or if it shall exist already.
            shell (bool): Whether to use a wrapper shell.

        """
        name = name or self.glowing()
        if not name and shell:
            execute(replace=True)
        elif not exists(self.path) and not new:
            raise ValueError('Orb storage folder "%s" does not exist' % self.path)
        elif not name:
            raise RuntimeError('There is no glowing orb')
        elif name not in self.orbs and not new:
            raise ValueError('Invalid orb name "%s"' % name)
        return Orb(name=name, orbs=self, shell=shell)


class Orb:
    def __init__(self, name, orbs, shell):
        """
        Args:
            name (str): The name of the orb.
            orbs (Orbs): An orb factory instance.
            shell (bool): Whether to activate the orb in a shell.

        """
        self.name = name
        self._orbs = orbs
        self._shell = shell

    def orb(self, shell_type=None):
        shell_type = shell_type or current_shell_type()
        return join(self._orbs.path, self.name, 'bin/activate_orb') + '.' + shell_type

    def make(self, reqs, executable=sys.executable, quiet=False, update=False):
        """
        Create or update the orb and generate a lockfile when necessary.

        Args:
            reqs (Requirements): The requirements to be used for package installation.
            executable (str): The Python executable to use.
            quiet (bool): Whether to suppress info messages.
            update (bool): Whether to update the orb including the relevant lockfile.

        """
        executable = realpath(which(executable))
        if not update and exists(reqs.locked) and reqs.changed:
            raise RuntimeError('Requirements lockfile of "%s" is outdated' % reqs.path)
        if not quiet:
            verb = 'Updating' if update else 'Making'
            print('%s orb "%s" using "%s"...' % (verb, self.name, reqs))
            print('Python executable: %s' % executable)

        # Creating virtual environment
        command = '%s -m venv --clear "%s"' % (executable, join(self._orbs.path, self.name))
        if execute(command=command).returncode:
            raise RuntimeError('Unable to create virtual environment')

        # Creating activation scripts
        for shell in SHELLS:
            activate = join(self._orbs.path, self.name, 'bin/activate')
            Path(self.orb(shell)).write_text(render('activate_orb.' + shell, {
                'name': self.name, 'cwd': getcwd(), 'init_file': self.orb(shell),
                'activate_script': activate + ('.' + shell if shell != 'bash' else ''),
            }))

        # Installing requirements
        command = 'source "%s" && pip install --upgrade pip wheel' % self.orb()
        command += ' && pip install --upgrade -r "%s"' % reqs
        if execute(command=command).returncode:
            raise RuntimeError('Unable to install requirements')

        # Generating lockfile
        if reqs.changed:
            freeze = 'pip freeze --all --exclude-editable | grep -v "pkg[-_]resources"'
            reqs.lock(self.activate(run=freeze, capture=True).stdout)
        if not quiet:
            print('Orb "%s" is ready for use' % self.name)

    def destroy(self):
        if environ.get('PYORBS_ACTIVE_ORB', None) == self.name:
            raise RuntimeError('You must exit the orb first for this operation')
        print('Destroying orb "%s"...' % self.name)
        if self._orbs.glowing() == self.name:
            self._orbs.toggle_glow(self.name)
        rmtree(join(self._orbs.path, self.name))

    def info(self):
        print('Orb "%s"' % self.name)
        old = self.activate(run='pip list --outdated', capture=True)
        print('\n' + (old.stdout.rstrip() if old.stdout else 'All packages are up-to-date'))

    def activate(self, run=None, no_cd=False, capture=False):
        """
        Activate the orb.

        Args:
            run (str): A command to run in the activated orb.
            no_cd (bool): Do not change the working directory after orb activation.
            capture (bool): Whether to capture the standard output and standard error.

        """
        if not exists(self.orb()):
            raise RuntimeError('Orb file "%s" not found' % self.orb())
        if not capture:
            print('Activating orb "%s"...' % self.name)
        if not run:
            self._orbs.toggle_glow(self.name, force_on=True)
        if run and not capture:
            print('Running "%s"...' % run)

        environ['PYORBS_SHELL'] = str(int(self._shell and not run))
        environ['PYORBS_NO_CD'] = str(int(no_cd))

        init = self.orb() if not run else None
        command = 'source "%s"; %s' % (self.orb(), run) if run else None
        return execute(init=init, command=command, replace=not capture, capture=capture)



import sys
from os import environ, execl
from subprocess import PIPE, run

DEFAULT_SHELL = '/bin/bash'
SHELLS = ('bash', 'fish')


def current_shell_type():
    for shell in SHELLS:
        if shell in environ.get('SHELL', DEFAULT_SHELL):
            return shell
    raise RuntimeError('Shell "%(SHELL)s" is not supported' % environ)


def execute(init=None, command=None, replace=False, capture=False):
    """
    Execute a shell command or start a new interactive session.

    Args:
        init (str): The init file to be used when starting an interactive session.
        command (str): The command to be executed.
        replace (bool): Whether to replace the current process or return the completed process.
        capture (bool): Whether to capture the standard output and standard error.

    """
    if init and command:
        raise ValueError('You cannot use an initialization file with a command')
    if replace and capture:
        raise ValueError('You cannot capture the output when replacing the current process')

    if current_shell_type() == 'bash':
        args = ['--init-file', init] if init else ['-c', command] if command else []
    elif current_shell_type() == 'fish':
        args = ['-C', 'source "%s"' % init] if init else ['-c', command] if command else []

    shell = environ.get('SHELL', DEFAULT_SHELL)
    sys.stdout.flush()
    sys.stderr.flush()
    if replace:
        return execl(shell, shell, *args)
    if capture:
        return run([shell] + args, stdout=PIPE, stderr=PIPE, universal_newlines=True, check=False)
    return run([shell] + args, check=False)


def which(command):
    """
    Returns the absolute path of the given command.
    """
    path = run(['which', command], stdout=PIPE, universal_newlines=True, check=False)
    if path.returncode:
        raise ValueError('Command "%s" not found' % command)
    return path.stdout.strip()



#!/usr/bin/env python3
import pathlib

import setuptools

from pyorbs import app

setuptools.setup(
    name=app.__project__,
    version=app.__version__,
    author=app.__author__,
    maintainer=app.__maintainer__,
    keywords=app.__keywords__,
    description=app.__doc__.strip(),
    long_description=pathlib.Path('README.rst').read_text(),
    long_description_content_type='text/x-rst',
    url=app.__urls__['docs'],
    project_urls={
        'Documentation': app.__urls__['docs'],
        'Issue tracker': app.__urls__['issues'],
        'Release notes': app.__urls__['releases'],
        'Source code': app.__urls__['source'],
    },
    python_requires='>=3.6',
    install_requires=[],
    packages=setuptools.find_packages(exclude=['docs*', 'tests*']),
    package_data={'pyorbs': ['templates/*']},
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={'console_scripts': ['orb=pyorbs.app:main']},
)



"""
A tool for managing Python virtual environments.
"""
import argparse
import os
import sys

from pyorbs.orbs import Orbs
from pyorbs.reqs import Requirements
from pyorbs.templates import render

__project__ = 'pyorbs'
__version__ = '1.6.1'
__author__ = 'Logikal'
__maintainer__ = 'Logikal'
__keywords__ = 'development virtual environment'
__urls__ = {
    'docs': 'https://pyorbs.readthedocs.io',
    'issues': 'https://github.com/logikal-code/pyorbs/issues',
    'releases': 'https://pyorbs.readthedocs.io/en/v%s/sections/changelog.html' % __version__,
    'source': 'https://github.com/logikal-code/pyorbs',
}


def main(args=tuple(sys.argv[1:])):  # pylint: disable=too-many-branches
    parser = argparse.ArgumentParser(description=__doc__)
    group = parser.add_argument_group(title='actions').add_mutually_exclusive_group()
    group.add_argument('-a', '--activate', action='store_true', help='activate an orb (default)')
    group.add_argument('-l', '--list', action='store_true', help='list orbs')
    group.add_argument('-m', '--make', action='store_true', help='make an orb')
    group.add_argument('-u', '--update', action='store_true', help='update an orb')
    group.add_argument('-d', '--destroy', action='store_true', help='destroy an orb')
    group.add_argument('-f', '--freeze', action='store_true', help='freeze requirements')
    group.add_argument('-t', '--test', action='store_true', help='test requirements')
    group.add_argument('-i', '--info', action='store_true', help='orb information')
    group.add_argument('-g', '--glow', action='store_true', help='toggle orb glow')
    group.add_argument('--bash', action='store_true', help='print bash completion script')
    parser.add_argument('name', nargs='?', help='name of the orb (default: glowing orb name)')
    parser.add_argument('-v', '--version', action='store_true', help='show version')
    parser.add_argument('-c', '--command', metavar='X',
                        help='command to be run after orb activation')
    parser.add_argument('-r', '--reqs', default='requirements.txt', metavar='X',
                        help='path to the requirements (default: requirements.txt)')
    parser.add_argument('-e', '--exec', default=sys.executable, metavar='X',
                        help='Python executable to use (default: sys.executable)')
    parser.add_argument('-o', '--orbs', default=os.path.expanduser('~/.pyorbs'), metavar='X',
                        help='orb storage directory (default: ~/.pyorbs)')
    parser.add_argument('-n', '--no-cd', action='store_true', help='do not change directory')
    parser.add_argument('-s', '--shell', action='store_true', help='activate the orb in a shell')
    parser.add_argument('-b', '--bare', action='store_true', help='use the bare requirements file')
    args = parser.parse_args(args)
    try:
        orbs = Orbs(path=args.orbs)
        if args.list:
            orbs.list()
        elif args.make:
            orbs.orb(args.name, new=True).make(Requirements(args.reqs, args.bare), args.exec)
        elif args.update:
            orbs.orb(args.name).make(Requirements(args.reqs, args.bare), args.exec, update=True)
        elif args.destroy:
            orbs.orb(args.name).destroy()
        elif args.info:
            orbs.orb(args.name).info()
        elif args.freeze:
            orbs.freeze(args.reqs, args.exec)
        elif args.test:
            return orbs.test(args.reqs)
        elif args.glow:
            orbs.toggle_glow(args.name)
        elif args.bash:
            print(render('orb-completion.bash').rstrip())
        elif args.version:
            print(__version__)
        else:
            orbs.orb(args.name, shell=args.shell).activate(run=args.command, no_cd=args.no_cd)
        return 0
    except KeyboardInterrupt:
        return 1
    except Exception as error:  # pylint: disable=broad-except
        return 'Error: %s' % error
